package de.mdelab.instanceGraphEditor.ui.model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Observable;

import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;

public class Node extends Observable {

	private Rectangle constraint;
	
	private String name;
	
	private EObject obj;
	
	private List<Link> sourceLinks, targetLinks;
	
	private Node parent;
	
	private List<Node> children = new ArrayList<Node>();
	
	private List<Node> eReferences = new ArrayList<Node>();
	
	private Graph graph;
	

	public void init() {
		
		setConstraint(new Rectangle(new Point(50, 50), new Dimension(100,100)));
		
		if (this.obj != null) {
			name = obj.eClass().getName();
			graph.addNode(this);
			
			//add referenced objects to the list of references
			for(EObject ref :obj.eCrossReferences()){
				List<EObject> references = graph.getReferences(ref);
				references.add(obj);
			}
			
			//create Containment links TO the child nodes, those which are already in the Graph.
			for (EObject child : obj.eContents()) {
				Node childNode = graph.nodeIsCreated(child);
				if (childNode != null) {
					children.add(childNode);
					this.createLink(this, childNode, getClassifierName(child));
				}
			}
			
			//create Links TO the parent node if parent already in the Graph
			if(obj.eContainer() != null && graph.nodeIsCreated(obj.eContainer()) != null){
				parent = graph.nodeIsCreated(obj.eContainer());
				parent.addChild(this);
				this.createLink(parent, this, parent.getClassifierName(obj));
			}
			
			// create Containment Links FROM the nodes which were expanded and contain the reference to this node
			// graph.getReferences(obj) returns the List of all object which registered as referencing the object
			List<EObject> references = graph.getReferences(obj);
			for(EObject ref : references){
				if(graph.nodeIsCreated(ref) != null){ // additional check for existence of objects in the Graph, since they might have been hidden, but will remain in the list.
					this.createLink(graph.nodeIsCreated(ref), this, graph.nodeIsCreated(ref).getClassifierName(obj));
				}
			}
			
			// create Reference Links TO the nodes which has already been created.
			for (EObject refObj : obj.eCrossReferences()) {
				Node existingNode = graph.nodeIsCreated(refObj);
				if (existingNode != null){
					this.createLink(this, existingNode, getClassifierName(existingNode.getEObject()));
					eReferences.add(existingNode);
				}
			}
		}
	}

	public void setGraph(Graph graph) {
		this.graph =  graph;
	}
	
	public Object getGraph() {
		return graph;
	}
	public void setEObjecct(EObject object) {
		this.obj =  object;
	}
	
	public Object getEObject() {
		return obj;
	}
	
	public void setParent(Node parent) {
		this.parent =  parent;
	}
	
	public Node getParent() {
		return parent;
	}
	
	public boolean hasChildren() {
		return (this.obj.eContents().size() != 0);
	}

	public EList<EAttribute> getAttributes() {
		if (obj != null) {
			EList<EAttribute> attr = obj.eClass().getEAllAttributes();
			return attr;
		}
		return new BasicEList<EAttribute>();
	}
	
	public EAttribute getAttributeByName(String attributeName){
		
		EList<EAttribute> attrs = obj.eClass().getEAllAttributes();
		System.out.println("attributeName: " + attributeName);
		for(EAttribute a: attrs)
			System.out.println("AttName: " + a.getName() );
		for(EAttribute a: attrs){
			if( a.getName().equals(attributeName)){
				return a;
			}
		}
		return null;
	}

	
	public void setAttributeValue(EAttribute attr, String newValue) {
		Object value = attr.getEAttributeType().getEPackage().getEFactoryInstance().createFromString((EDataType) attr.getEType(), newValue);
		obj.eSet(attr, value);
		setChanged();
		notifyObservers();
	}

	public Object getAttributeValue(EAttribute attr) {
		return obj.eGet(attr);
	}
	
	public void expand() {
		
		//expand parent if has not yet been expanded 
		if (obj.eContainer() != null && graph.nodeIsCreated(obj.eContainer()) == null) {
			parent = InstanceGraphFactory.INSTANCE.createNode(graph, obj.eContainer(), null);
		}
		
		//expand children. those which has not been opened
		for (EObject child : obj.eContents()) {
			if (graph.nodeIsCreated(child) == null) {
				Node node = InstanceGraphFactory.INSTANCE.createNode(graph, child, this);
				children.add(node);
			}
		}
		
		//expand cross-references, also check whether the node already in the graph
		for (EObject refObj : obj.eCrossReferences()) {
			Node existingNode = graph.nodeIsCreated(refObj);
			if (existingNode == null){
				Node node = InstanceGraphFactory.INSTANCE.createNode(graph, refObj, this);
				eReferences.add(node);
			}else if(existingNode == this){
				this.createLink(this, existingNode, getClassifierName(getClassifierName(existingNode.getEObject())));
			}
		}
	}
	
	public void hideNode(){
		
		//remove incoming link to the node (Parent node!)
		List<Link> trgC = getTargetConnections();
		if (!trgC.isEmpty()) {
			Link trg = ((Link) trgC.get(0));
			trg.getSource().removeSourceLink(trg);
			removeTargetLink(trg);
		}
		
		//remove EReference links, which are left after removing children Links
		removeAllSourceLinks();
		removeAllTargetLinks();
		
		graph.removeNode(this);
		parent.removeChild(this);
	}
	 

	public void hideChildren() {

		for (Node n : children) {
			n.hideChildren();
			// remove all incoming link to the child node
			n.removeAllTargetLinks();

		}
		graph.removeNodes(children);
	}
	
	public void hideCrossReferences() {

		for (Node n : eReferences) {
			n.hideChildren();
			// remove all incoming link to the reference node
			n.removeAllTargetLinks();
		}
		graph.removeNodes(eReferences);
	}	


	
	public void addChild(Node child){
		children.add(child);
	}
	
	public void removeChild(Node child){
		children.remove(child);
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public Rectangle getConstraint() {
		return constraint;
	}
	
	public void setConstraint(Rectangle constraint) {
		this.constraint = constraint;
		setChanged();
		notifyObservers();
	}
	
	public String getClassifierName(Object referent){
		
		List<EReference> refList = obj.eClass().getEAllReferences();
		for (EReference ref : refList){
			if(ref.isMany()){
				@SuppressWarnings("unchecked")
				Collection<EObject> objs =  (Collection<EObject>)obj.eGet(ref);
					if (objs.contains(referent)){
						return ref.getName();
				}
			}else if(obj.eGet(ref) == referent){
					return ref.getName();
			}
		}
		return null;
	}
	
	private void createLink(Node source, Node target, String linkLable){
		Link c = new Link(linkLable);
		c.setSource(source);
		c.setTarget(target);
	}

	public List<Link> getSourceLinks() {
		if (sourceLinks == null) {
			sourceLinks = new ArrayList<Link>();
		}
		return sourceLinks;
	}

	public List<Link> getTargetConnections() {
		if (targetLinks == null) {
			targetLinks = new ArrayList<Link>();
		}
		return targetLinks;
	}

	public void addSourceLink(Link link) {
		getSourceLinks().add(link);
		setChanged();
		notifyObservers();
	}

	public void addTargetLink(Link link) {
		getTargetConnections().add(link);
		setChanged();
		notifyObservers();
	}

	public void removeSourceLink(Link link) {
		getSourceLinks().remove(link);
		setChanged();
		notifyObservers();
	}
	
	public void removeAllSourceLinks(){
		//delete must be called on the link; removes links from the list of links in the target object and source object
		while (!getSourceLinks().isEmpty())
			getSourceLinks().get(0).delete();
	}
	
	public void removeAllTargetLinks(){
		//delete must be called on the link; removes links from the list of links in the target object and source object
		while (!getTargetConnections().isEmpty())
			getTargetConnections().get(0).delete();
	}

	public void removeTargetLink(Link link) {
		getTargetConnections().remove(link);
		setChanged();
		notifyObservers();
	}
	
	public void printout(){
		
		System.out.println("Node name  :" + name +", EObject: " + obj);
		System.out.println("eContents()");
		for(Node n: children){
			System.out.println("node name: " + n.name +", nodeExist? " + graph.nodeIsCreated(n) +", EObject: " + n.obj);
		}
		System.out.println("");
		System.out.println("eCrossReference()");
		for(Node n: eReferences){
			System.out.println("node name: " + n.name +", nodeExist? " + graph.nodeIsCreated(n) +", EObject: " + n.obj);
		}
		System.out.println("");
	}
}