package de.mdelab.sdm.interpreter.code.debug.ui.launcher;

import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.debug.core.ILaunch;
import org.eclipse.debug.core.ILaunchConfiguration;
import org.eclipse.debug.core.ILaunchManager;
import org.eclipse.debug.core.model.IDebugTarget;
import org.eclipse.debug.core.model.LaunchConfigurationDelegate;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.ui.console.ConsolePlugin;
import org.eclipse.ui.console.IConsole;
import org.eclipse.ui.console.IConsoleManager;
import org.eclipse.ui.console.MessageConsole;
import org.eclipse.ui.console.MessageConsoleStream;

import de.mdelab.resourceSetSynchronizer.ConnectionUtil;
import de.mdelab.sdm.interpreter.core.variables.Variable;

public abstract class SDInterpreterLaunchDelegate<Activity extends EObject, Classifier> extends LaunchConfigurationDelegate {
	
	private SDServerThread<Activity, Classifier> launchInterpreter( final ILaunchConfiguration p_configuration,
																	final Integer p_serverPort,
																	final Integer p_resourceSetSynchronizerPort,
																	final IProgressMonitor p_monitor ) 
	throws CoreException {
		final ResourceSet resSet = new ResourceSetImpl();
		final String sdUriStr = p_configuration.getAttribute( ISDLaunchUIConstants.ATTR_STORY_DIAGRAM_URI, (String) null );
		final Resource diagRes = resSet.getResource( URI.createURI( sdUriStr ), true );
		
		final int selectedActivityIndex = p_configuration.getAttribute(	ISDLaunchUIConstants.ATTR_STORY_ACTIVITY_SELECTION_INDEX, -1 );
		@SuppressWarnings("unchecked")
		final Activity activity = (Activity) diagRes.getContents().get( selectedActivityIndex );

		final List<Variable<Classifier>> variableBindings = new ArrayList<Variable<Classifier>>();
		final Map<String, String> parameters = p_configuration.getAttribute( ISDLaunchUIConstants.ATTR_STORY_ACTIVITY_PARAMETERS, Collections.<String, String> emptyMap() );
		
		for ( final Map.Entry<String, String> entry : parameters.entrySet() ) {
			final EObject eObject = resSet.getEObject( URI.createURI( entry.getValue() ), true );
			final Variable<Classifier> variable = createVariable( entry.getKey(), eObject );
			variableBindings.add( variable );
		}

		final IConsoleManager consoleManager = ConsolePlugin.getDefault().getConsoleManager();
		final MessageConsole console = new MessageConsole( "Story Diagram Interpreter Notifications", null );
		final MessageConsoleStream consoleStream = new MessageConsoleStream( console );
		consoleManager.addConsoles( new IConsole[]{ console } );
		consoleManager.showConsoleView( console );

		final Map<String, String> instanceModelUriStr = p_configuration.getAttribute( ISDLaunchUIConstants.ATTR_INSTANCE_MODEL_URI, Collections.<String, String> emptyMap() );
		final Map<Resource, URI> instanceModels = new HashMap<Resource, URI>();
		
		for ( final Map.Entry<String, String> entry : instanceModelUriStr.entrySet() ) {
			final Resource inputRes = resSet.getResource( URI.createURI( entry.getKey() ), true );
			final String outUriStr = entry.getValue();
			final URI outUri = outUriStr == null || outUriStr.isEmpty() ? null : URI.createURI( outUriStr );
			instanceModels.put( inputRes, outUri );
		}
 
		final boolean logOutput = p_configuration.getAttribute( ISDLaunchUIConstants.ATTR_LOG_OUTPUT, false );
		final String execTraceUriStr = p_configuration.getAttribute( ISDLaunchUIConstants.ATTR_EXECUTION_TRACE_URI, "" );
			
		final SDServerThread<Activity, Classifier> serverThread = createServerThread(	p_configuration,
																						activity,
																						variableBindings,
																						p_serverPort,
																						p_resourceSetSynchronizerPort,
																						p_monitor,
																						consoleStream,
																						instanceModels,
																						logOutput,
																						execTraceUriStr );

		if ( p_serverPort == null ) {
			serverThread.run();
		}
		else {
			serverThread.start();
		}
		
		return serverThread;
	}
	
	protected abstract Variable<Classifier> createVariable( String p_varName,
															EObject p_varValue );
	
	protected abstract SDServerThread<Activity, Classifier> createServerThread(	final ILaunchConfiguration p_configuration,
																				Activity p_activity,
																				List<Variable<Classifier>> p_variableBindings,
																				Integer p_serverPort,
																				Integer p_resourceSetSynchronizerPort,
																				IProgressMonitor p_monitor,
																				OutputStream p_messageStream,
																				Map<Resource, URI> p_instanceModelsr,
																				boolean pb_logOutput,
																				String p_execTraceUriStr);
	
	public void launch(	final ILaunchConfiguration p_configuration,
						final String p_mode,
						final ILaunch p_launch,
						final IProgressMonitor p_monitor )
	throws CoreException {
		p_monitor.beginTask( "Launching Story Diagram Interpreter...", 1 );
		
		try {
			if ( ILaunchManager.RUN_MODE.equals( p_mode ) ) {
				launchRun( p_configuration, p_launch, p_monitor );
			}
			else if ( ILaunchManager.DEBUG_MODE.equals( p_mode ) ) {
				launchDebug( p_configuration, p_launch, p_monitor );
			}
			else {
				throw new CoreException( new Status( IStatus.ERROR, ISDLaunchUIConstants.PLUGIN_ID, "Unknown execution mode " + p_mode ) );
			}
		}
		finally {
			p_monitor.worked( 1 );
			p_monitor.done();			
		}
	}
	
	private void launchDebug( 	final ILaunchConfiguration p_configuration,
								final ILaunch p_launch,
								final IProgressMonitor p_monitor ) 
	throws CoreException {
		final boolean useInterpreterServer = p_configuration.getAttribute( ISDLaunchUIConstants.ATTR_USE_SERVER_INTERPRETER, false );

		final Integer serverPort;
		final Integer lastUsedPort;
		final String serverAddress;
		
		if ( useInterpreterServer ) {
			serverPort = Integer.parseInt( p_configuration.getAttribute( ISDLaunchUIConstants.ATTR_SERVER_PORT, "" ) );
			serverAddress = p_configuration.getAttribute( ISDLaunchUIConstants.ATTR_SERVER_ADDRESS, "" );
			lastUsedPort = serverPort;
		}
		else {
			serverPort = ConnectionUtil.findNextAvailablePort( 1024 );
			serverAddress = ISDLaunchUIConstants.SERVER_DEFAULT_ADDRESS;
			lastUsedPort = ConnectionUtil.findNextAvailablePort( serverPort );
			
			final SDServerThread<?, ?> serverThread = launchInterpreter( p_configuration, serverPort, lastUsedPort, p_monitor );

			try {
				while ( !serverThread.isReady() ) {
					Thread.sleep( 100 );
				}
			} 
			catch ( final InterruptedException p_ex) {
				p_ex.printStackTrace();
			}
		}

		final Integer clientPort = ConnectionUtil.findNextAvailablePort( lastUsedPort );
		final IDebugTarget target = createDebugTarget( p_launch, serverAddress, serverPort, clientPort, ConnectionUtil.findNextAvailablePort( clientPort ) );
		p_launch.addDebugTarget( target );
	}
	
	protected abstract IDebugTarget createDebugTarget( 	ILaunch p_launch,
														String p_serverAddress,
														int pi_serverPort,
														int pi_clientPort,
														int pi_resSetSyncPort )
	throws CoreException;
	
	protected abstract IDebugTarget createRunTarget( ILaunch p_launch );

	private void launchRun( final ILaunchConfiguration p_configuration,
							final ILaunch p_launch,
							final IProgressMonitor p_monitor ) 
	throws CoreException {
		final IDebugTarget target = createRunTarget( p_launch );
		p_launch.addDebugTarget( target );
		
		try {
			launchInterpreter( p_configuration, null, null, p_monitor );
		}
		finally {
			target.terminate();		
		}
	}
}
