package de.mdelab.sdm.interpreter.core.debug.client;

import java.util.Stack;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.net4j.Net4jUtil;
import org.eclipse.net4j.connector.IConnector;
import org.eclipse.net4j.signal.SignalProtocol;
import org.eclipse.net4j.tcp.ITCPAcceptor;
import org.eclipse.net4j.tcp.TCPUtil;
import org.eclipse.net4j.util.container.ContainerUtil;
import org.eclipse.net4j.util.container.IManagedContainer;
import org.eclipse.net4j.util.lifecycle.LifecycleUtil;
import org.eclipse.spi.net4j.ServerProtocolFactory;

import de.mdelab.resourceSetSynchronizer.ConnectionUtil;
import de.mdelab.resourceSetSynchronizer.EObjectUuidCalculator;
import de.mdelab.resourceSetSynchronizer.SynchronizerAdapter;
import de.mdelab.sdm.interpreter.core.debug.client.protocol.SDDebugProtocolEventIndicationFactory;
import de.mdelab.sdm.interpreter.core.debug.client.protocol.requests.ConnectClientCommandRequest;
import de.mdelab.sdm.interpreter.core.debug.client.protocol.requests.GetExecutionStackCommandRequest;
import de.mdelab.sdm.interpreter.core.debug.client.protocol.requests.RemoveBreakpointRequest;
import de.mdelab.sdm.interpreter.core.debug.client.protocol.requests.ResumeCommandRequest;
import de.mdelab.sdm.interpreter.core.debug.client.protocol.requests.SetBreakpointCommandRequest;
import de.mdelab.sdm.interpreter.core.debug.client.protocol.requests.SetVariableCommandRequest;
import de.mdelab.sdm.interpreter.core.debug.client.protocol.requests.StepBackOverRequest;
import de.mdelab.sdm.interpreter.core.debug.client.protocol.requests.StepIntoCommandRequest;
import de.mdelab.sdm.interpreter.core.debug.client.protocol.requests.StepOutCommandRequest;
import de.mdelab.sdm.interpreter.core.debug.client.protocol.requests.StepOverCommandRequest;
import de.mdelab.sdm.interpreter.core.debug.client.protocol.requests.SuspendCommandRequest;
import de.mdelab.sdm.interpreter.core.debug.client.protocol.requests.TerminateCommandRequest;
import de.mdelab.sdm.interpreter.core.debug.client.stack.DebugClientStackFrame;
import de.mdelab.sdm.interpreter.core.debug.protocol.SDDebugProtocolCommandIndicationFactory;
import de.mdelab.sdm.interpreter.core.notifications.NotificationTypeEnum;

public abstract class SDDebugClient<StoryDiagramElement extends EObject, Activity> {
	
	private final int					debugClientPort;

	private IManagedContainer			debugProtocolContainer;

	private SignalProtocol<Object>		debugProtocol;

	private BlockingQueue<DebugEvent>	debugEventQueue;

	private ITCPAcceptor				acceptor;

	private ResourceSetImpl				debuggerResourceSet;

	private SynchronizerAdapter			synchronizerAdapter;

	private int							synchronizerAdapterPort;

	public SDDebugClient( 	final int debugClientPort,
							final int pi_synchronizerAdapterPort ) {
		this.debugClientPort = debugClientPort;
		synchronizerAdapterPort = pi_synchronizerAdapterPort;
	}
//
//	public void connect( final String debugServerAddress ) 
//	throws Exception {
//		connect( debugServerAddress, debugClientPort );//MLSDMDebugger.MLSDM_DEBUG_SERVER_DEFAULT_PORT );
//	}

	@SuppressWarnings("unchecked")
	public synchronized void connect(String debugServerAddress, int debugServerPort) 
	throws Exception {
		assert this.debugProtocolContainer == null;

		assert debugServerAddress != null;
		assert !"".equals(debugServerAddress);

		this.debugEventQueue = new LinkedBlockingQueue<DebugEvent>();

		/*
		 * Open debug client server to receive events from the debugger.
		 */
		this.debugProtocolContainer = ContainerUtil.createContainer();

		Net4jUtil.prepareContainer(this.debugProtocolContainer);
		TCPUtil.prepareContainer(this.debugProtocolContainer);
		this.debugProtocolContainer.registerFactory(new SDDebugProtocolEventIndicationFactory<StoryDiagramElement>(this));

		LifecycleUtil.activate(this.debugProtocolContainer);

		this.acceptor = TCPUtil.getAcceptor(this.debugProtocolContainer, "0.0.0.0:" + this.debugClientPort);

		/*
		 * Create debug protocol
		 */
		debugProtocol = (SignalProtocol<Object>) debugProtocolContainer.getFactory( ServerProtocolFactory.PRODUCT_GROUP, SDDebugProtocolCommandIndicationFactory.SDE_DEBUG_PROTOCOL ).create( null );

		/*
		 * Open connection to debug server
		 */
		try {
			final IConnector connector = TCPUtil.getConnector( debugProtocolContainer, debugServerAddress + ":" + debugServerPort );
		
	        // DB: The default timeout value seems to be too small. Regularly a timeout occurs before the connection has time to establish
			connector.setOpenChannelTimeout( ConnectionUtil.DEFAULT_CONNECTION_TIMEOUT );
			this.debugProtocol.open(connector);
	
			/*
			 * Send connect message
			 */
			final int resourceSetSynchronizerPort = new ConnectClientCommandRequest<StoryDiagramElement>( debugProtocol, this, debugClientPort ).send();
	
			/*
			 * Create a resource set
			 */
			this.debuggerResourceSet = new ResourceSetImpl();
	
			this.synchronizerAdapter = new SynchronizerAdapter( this.debuggerResourceSet, synchronizerAdapterPort );
			this.synchronizerAdapter.getEObjectUUIDManager().setEObjectUuidCalculator( getStoryDiagramElementUuidCalculator() );
	
			this.synchronizerAdapter.connect( debugServerAddress, resourceSetSynchronizerPort );
		}
		catch ( final Exception p_ex ) {
			// DB: We must disconnect in case of exception (e.g. the server was not running) otherwise the attempts to 
			// connect again will never work and the application will need to be restarted
			disconnect( false );
			
			throw p_ex;
		}
	}

	protected abstract EObjectUuidCalculator getStoryDiagramElementUuidCalculator();

	public synchronized void disconnect( final boolean pb_deconnectRemote ) {		
		System.out.println("Shutting down client.");

		// DB: Release all resources that were created even if the debug protocol was not
		if ( acceptor != null ) {
			LifecycleUtil.deactivateNoisy( acceptor );
		}
		
		if ( debugProtocolContainer != null ) {
			LifecycleUtil.deactivateNoisy( debugProtocolContainer );
		}
		
		if ( debugProtocol != null && debugProtocol.isActive() ) {

			try {
				synchronizerAdapter.shutdown( pb_deconnectRemote );

				System.out.println("Client shut down.");

				assert !this.debugProtocol.isActive();
				assert !this.debugProtocolContainer.isActive();
			}
			catch ( final RuntimeException p_ex )	{
				throw p_ex;
			}
			catch ( final Exception p_ex )	{
				throw new RuntimeException( "Could not shutdown synchronizer adapter.", p_ex );
			}
		}
	}

	public BlockingQueue<DebugEvent> getDebugEventQueue() {
		return debugEventQueue;
	}

	/**
	 * Sends a resume command to the debugger.
	 * 
	 * @throws Exception
	 */
	public void resume() throws Exception
	{
		Exception ex = new ResumeCommandRequest<StoryDiagramElement>(this.debugProtocol, this).send();

		if (ex != null)
		{
			throw ex;
		}
	}

	/**
	 * Sends a suspend command to the debugger
	 * 
	 * @throws Exception
	 */
	public void suspend() throws Exception
	{
		Exception ex = new SuspendCommandRequest<StoryDiagramElement>(this.debugProtocol, this).send();

		if (ex != null)
		{
			throw ex;
		}
	}

	/**
	 * Sends a terminate command to the debugger.
	 * 
	 * @throws Exception
	 */
	public void terminate() 
	throws Exception {
		final Exception ex = new TerminateCommandRequest<StoryDiagramElement>( debugProtocol, this ).send();
		
		// DB: Ensure the wait for event queue will terminate
		getDebugEventQueue().add( new ExecutionTerminatedEvent() );

		if ( ex != null ) {
			throw ex;
		}
	}

	/**
	 * Sends a step into command to the debugger.
	 * 
	 * @throws Exception
	 */
	public void stepInto() throws Exception
	{
		Exception ex = new StepIntoCommandRequest<StoryDiagramElement>(this.debugProtocol, this).send();

		if (ex != null)
		{
			throw ex;
		}
	}

	/**
	 * Sends a step over command to the debugger.
	 * 
	 * @throws Exception
	 */
	public void stepOver() throws Exception
	{
		Exception ex = new StepOverCommandRequest<StoryDiagramElement>(this.debugProtocol, this).send();

		if (ex != null)
		{
			throw ex;
		}
	}

	/**
	 * Sends a step out command to the debugger.
	 * 
	 * @throws Exception
	 */
	public void stepOut() throws Exception
	{
		Exception ex = new StepOutCommandRequest<StoryDiagramElement>(this.debugProtocol, this).send();

		if (ex != null)
		{
			throw ex;
		}
	}

	public void stepBackOver() throws Exception
	{
		Exception ex = new StepBackOverRequest<StoryDiagramElement>(this.debugProtocol, this).send();

		if (ex != null)
		{
			throw ex;
		}
	}

	public String setBreakpoint(String elementUUID) throws Exception
	{
		return this.setBreakpoint(elementUUID, null, null);
	}

	public String setBreakpoint(String elementUUID, String expressionLanguage, String expression)
	throws Exception {
		return new SetBreakpointCommandRequest<StoryDiagramElement>(this.debugProtocol, this, elementUUID, expressionLanguage, expression).send();
	}

	/**
	 * Waits for the next debug event of the specified type, or simply the next
	 * debug event if evenType is null.
	 * 
	 * @param eventType
	 * @return
	 * @throws InterruptedException
	 */
	public DebugEvent waitForDebugEvent(Class<? extends DebugEvent> eventType)
	throws InterruptedException	{
		DebugEvent debugEvent;

		do {
			debugEvent = this.getDebugEventQueue().take();
			System.out.println("DEBUG CLIENT: Received event: " + debugEvent);
		}
		while (eventType != null && !eventType.isInstance(debugEvent) && !(debugEvent instanceof ExecutionFinishedEvent)
				&& !(debugEvent instanceof ExecutionTerminatedEvent));

		System.out.println("DEBUG CLIENT: Returning event: " + debugEvent);

		return debugEvent;
	}

	public ResourceSet getDebuggerResourceSet()
	{
		return this.debuggerResourceSet;
	}

	public SynchronizerAdapter getSynchronizerAdapter()
	{
		return this.synchronizerAdapter;
	}

	public Stack<DebugClientStackFrame<StoryDiagramElement>> getExecutionStack() throws Exception
	{
		return new GetExecutionStackCommandRequest<StoryDiagramElement>(this.debugProtocol, this).send();
	}

	public void setVariable(DebugClientStackFrame<StoryDiagramElement> stackFrame, String variableName, Object newValue) throws Exception
	{
		Exception ex = new SetVariableCommandRequest<StoryDiagramElement>(this.debugProtocol, this, variableName, stackFrame, newValue)
				.send();

		if (ex != null)
		{
			throw ex;
		}
	}

	public void removeBreakpoint(String breakpointUUID) throws Exception
	{
		Exception ex = new RemoveBreakpointRequest<StoryDiagramElement>(this.debugProtocol, this, breakpointUUID).send();

		if (ex != null)
		{
			throw ex;
		}
	}

	@SuppressWarnings("unchecked")
	public Activity getExecutedActivity() throws Exception
	{
		Stack<DebugClientStackFrame<StoryDiagramElement>> stack = this.getExecutionStack();

		for (int i = stack.size() - 1; i >= 0; i--)
		{
			DebugClientStackFrame<StoryDiagramElement> stackFrame = stack.get(i);

			if (stackFrame.getLastNotificationType() == NotificationTypeEnum.ACTIVITY_EXECUTION_STARTED
					|| stackFrame.getLastNotificationType() == NotificationTypeEnum.ACTIVITY_EXECUTION_FINISHED)
			{
				return (Activity) stackFrame.getStoryDiagramElement();
			}
		}

		return null;
	}

	public StoryDiagramElement getCurrentStoryDiagramElement() throws Exception
	{
		return this.getExecutionStack().peek().getStoryDiagramElement();
	}

	public boolean isConnected() {
		return this.debugProtocol != null && this.debugProtocol.isActive();
	}
}
